/**
  * @file compute_IMAS_matches.h
  * @author Mariano Rodríguez
  * @date 2017
  * @brief Computing IMAS matches among generalised keypoints as in \cite imas_IPOL_2017 .
  * @warning This file migth be linked to the patent Jean-Michel Morel and Guoshen Yu, Method and device for the invariant affine recognition recognition of shapes (WO/2009/150361), patent pending.
  */

#ifndef MATCHING_H
#define MATCHING_H

#include <math.h>
#include <algorithm>
#include <ctime>
#include <cstdlib>
#include <vector>
#include <cstdio>
#include "sift.h"
#include "stats/stats.hpp"

#define FILTER_RADIUS 4

/************* DATATYPES ********************/

/**
 * @brief Definition of a keypoint
 */
struct KeyPoint
{
    float x;
    float y;
    float scale;
    float angle;
    void* kp_ptr;
};

/**
 * @brief Used to aggregate spatially keypoints
 */
struct KeyPoints
{
    float x, y, sum_x, sum_y;
    std::vector<KeyPoint> KPvec;
};

/**
 * @brief Keypointlist is used to store keypoints during the computation
 */
typedef std::vector<KeyPoint> Keypointlist;

/**
 * @brief A simple version of a keypoint (without descriptor).
 */
struct Keypoint_simple {
    float x, y, scale, angle;
};

/**
 * @brief The very definition of a match as pair of keypoints.
 */
typedef std::pair<Keypoint_simple, Keypoint_simple> Matching;

/**
 * @brief List of matches.
 */
typedef std::vector<Matching> Matchingslist;

/********************************************/
/************* FUNCTIONS ********************/

void perform_matching(int channels, std::vector<float>& image, int w, int h, std::vector<float>& data, Matchingslist& matchings, int ps, float tau, bool automatic);

int compute_matches(int c, std::vector<KeyPoint *> &keys, Matchingslist &matchings, int ps, float tau, bool automatic);

/**
 * @brief Computes all generalised keypoints.
 * @param image: Input image.
 * @param width: Width of the input image.
 * @param height: Height of the input image.
 * @param keys: Returns a list of generalised keypoints.
 * @param stats A vector with statistics on found generalised keypoints. Mean, min or max of SIIM keypoints over all found generalised keypoints.
 * @return The total number of generalised keypoints that have been found.
 */
int compute_keypoints(std::vector<float>& image, int width, int height, int channels, std::vector<KeyPoint *> &keys, std::vector<float> &stats);


#endif // MATCHING_H
